#!/usr/bin/python
#  -*- coding: utf-8 -*-

from __future__ import absolute_import

import json
import zipfile

from PIL import Image
from StringIO import StringIO as io
from collections import defaultdict

from MDmisc.DefaultOrderedDict import DefaultOrderedDict
from NIST import NISTf
from NIST.fingerprint.functions import lstTo012

from .release import Database

class DatabaseExtended( Database ):
    def export_exercise( self, exercise = None, username = None ):
        sql = '''
            SELECT
                exercise.name AS exname,
                person.username AS username,
                exercise.mark AS mark_image,
                exercise.print AS print_image,
                minutia.x, minutia.y, minutia.theta
            
            FROM assignment
            
                JOIN object ON assignment.id = object.id 
                JOIN person ON object.creator = person.id
                JOIN exercise ON exercise.id = assignment.exercise
                JOIN minutia ON assignment.id = minutia.assignment
                
            WHERE
            
                assignment.ispreview IS false'''
        
        if exercise:
            sql += " AND exercise.name SIMILAR TO '%s'" % exercise
        
        if username:
            sql += " AND person.username SIMILAR TO '%s'" % username
        
        c = self.conn.cursor()
        res = c.execute( sql )
        
        return c
        
    def getImage( self, id ):
        return Image.open( io( self.query( "select data from image where id = '%s' LIMIT 1" % id ).fetchone()['data'] ) )
    
    def getImageInformations( self, id ):
        return self.query( "SELECT image.width, image.height, image.resolution FROM public.image WHERE image.id = '%s';" % id ).fetchone()
    
    def exportExerciseData( self, exercisename = None, username = None, limit = None ):
        sql = """
            SELECT *
            FROM
            (
                SELECT
                    row_number() over ( PARTITION BY exercise.name, person.username ) as row_num,
                    exercise.name AS exercisename,
                    exercise.mark,
                    exercise.print,
                    person.username,
                    person.name as personname,
                    person.id as userid, 
                    resulttype.name AS type, 
                    resultstep.name AS step, 
                    result.data
                
                FROM 
                    public.assignment, 
                    public.exercise, 
                    public.result, 
                    public.object, 
                    public.person, 
                    public.resulttype, 
                    public.resultstep
                
                WHERE 
                    assignment.exercise = exercise.id AND
                    assignment.id = object.id AND
                    result.assignment = assignment.id AND
                    result.type = resulttype.id AND
                    result.step = resultstep.id AND
                    object.creator = person.id AND
                    resultstep.name = 'Comparison'
                
                ORDER BY
                    exercise.name ASC,
                    person.username ASC,
                    result.type ASC, 
                    result.step ASC
            ) AS data
            
            WHERE data.row_num = 1
        
        """
        
        if exercisename != None:
            sql += " AND exercisename = '%s'" % exercisename
        
        if username != None:
            sql += " AND ( username = '%s' OR personname = '%s' )" % ( username, username ) 
        
        if limit != None:
            sql += " LIMIT %s" % limit
            
        return self.query( sql )

    def txt2Storage( self, data ):
        j = json.loads( data )
        
        rdefaultdict = lambda: defaultdict( rdefaultdict )
        
        userdata = rdefaultdict()
        
        for t in [ 'mark', 'print' ]:
            for key, value in j[ t ]['users'].items():
                for m, d in value['minutiae'].items():
                    userdata[t]['minutiae'][m] = d
            
            for key, value in j[ t ]['image'].items():
                userdata[t]['image'][ key ] = value
        
        userdata[ 'data' ] = data
        
        annotations = myStorage()
        annotations.loadFromJSON( userdata )
        annotations.pairing()
        
        return annotations
    
    def storage2ZIP( self, storage, exname, outres = "Pianos", crop = False ):
        
        ################################################################################
        #    Buffers
        ################################################################################
        
        zipbuf = io()
        tmpbuf = io()
    
        with zipfile.ZipFile( zipbuf, mode = "w", compression = zipfile.ZIP_DEFLATED ) as zf:
            
            ############################################################################
            #    Input file
            ############################################################################
            
            zf.writestr( "input.txt", storage.j[ 'data' ] )
            
            ############################################################################
            #    NIST data
            ############################################################################
             
            for imgtype in [ 'print', 'mark' ]:
                imgid = storage.j[imgtype]['image']['href'].replace( "?action=get_image&id=", "" )
                
                nistfile = NISTf()
                nistfile.add_Type01()
                nistfile.add_Type02()
                nistfile.add_Type09( None, 1 )
                nistfile.add_Type13( ( 500, 500 ), 500, 1 )
                
                nistfile.set_field( "2.007", exname )
                nistfile.set_field( "13.925", "Pianos" )
                
                ############################################################
                #    Image
                ############################################################
                
                w, h, res = self.query( "SELECT image.width, image.height, image.resolution FROM public.image WHERE image.id = '%s';" % imgid ).fetchone()
                img = self.getImage( imgid )
                
                nistfile.set_latent( img, res, 1 )
                  
                ################################################################
                #    Minutiae
                ################################################################
                
                i = 0
                minu = storage.filter( "imgtype", imgtype )
                              
                field008 = None
                field012 = []
                pairings = []
                 
                for minindex, minvalue in minu.items():
                    x, y = nistfile.cooPIL2NIST( tuple( minvalue['pos'] ) )
                    x = int( 100 * x )
                    y = int( 100 * y )
                    
                    theta = ( -minvalue['angle'] + 180 ) % 360
                    mintype = minvalue['type']
                      
                    if mintype == "center":
                        nistfile.set_field( "9.008", "%04d%04d" % ( x, y ) )
                      
                    else:
                        i += 1
                          
                        if mintype == "ridge_ending":
                            mintype = "A"
                        elif mintype == "bifurcation":
                            mintype = "B"
                        else:
                            mintype = "D"
                        
                        field012.append( 
                            [
                                "%03d" % i,
                                "%04d%04d%03d" % ( x, y, theta ),
                                "%02d" % 0,
                                "%s" % mintype
                            ]
                        )
                        
                        try:
                            pairings.append( "%s:%s" % ( i, minvalue['pairing'] ) )
                        except:
                            pass
                        
                nistfile.set_field( "9.010", i ) 
                nistfile.set_field( "9.012", lstTo012( field012 ) )
                
                ########################################################################
                #    Change resolution
                ########################################################################
                
                if outres != None and outres != "Pianos":
                    nistfile.changeResolution( outres )
                
                ########################################################################
                #    File Cropping
                ########################################################################
                
                if crop == "1":
                    inch = nistfile.get_resolution()
                    nistfile.crop( ( inch, inch ), centermm = nistfile.get_center() )
                
                ########################################################################
                #    files preparation for ZIP
                ########################################################################
                
                zf.writestr( "%s-pairing.txt" % imgtype, ", ".join( pairings ) )
                
                zf.writestr( imgtype + ".nist", nistfile.dump() )
                
                nistfile.getType13Annotated().save( tmpbuf, format = "JPEG" )
                zf.writestr( imgtype + ".jpeg", tmpbuf.getvalue() )
                tmpbuf.truncate( 0 )
            
        return zipbuf
    
class myStorage:
    def __init__( self ):
        self.data = DefaultOrderedDict( dict )
        self.j = None

    def loadFromString( self, data ):
        self.j = json.loads( data )
        self.processJson()
    
    def loadFromJSON( self, data ):
        self.j = data
        self.processJson()
    
    def processJson( self ):
        i = 1
        
        for type in self.j:
            if not type in [ 'print', 'mark' ]:
                continue
            
            try:
                for min in self.j[type]['minutiae']:
                    tmp = { 'imgtype': type }
                    for carac in self.j[type]['minutiae'][min]:
                        tmp[carac] = self.j[type]['minutiae'][min][carac]
    
                    self.data[ i ] = tmp
                    i += 1
            except:
                print "no minutiae for " + type
    
    def pairing( self ):
        lst = []
        for key, value in self.data.items():
            try:
                val = value['pairing']
            except:
                val = 0
                
            lst.append( ( key, val ) )
            
        lst = sorted( lst, key = lambda x: x[1] )
        
        tmp = DefaultOrderedDict()
        i = 1
        for key, value in lst:
            tmp[ i ] = self.data[ key ]
            i += 1
        
        self.data = tmp
        
    def pprint( self ):
        for d in self.data.iteritems():
            print d

    def list( self, field ):
        tmp = []

        for d in self.data.iteritems():
            try:
                tmp.append( d[ 1 ][ field ] )
            except:
                continue

        return tmp

    def filter( self, field, value = "", opp = "==" ):
        tmp = DefaultOrderedDict( dict )

        for d in self.data.iteritems():
            if field == "pairing":
                try:
                    if d[ 1 ][ field ]:
                        tmp[ d[ 0 ] ] = d[ 1 ]
                except:
                    continue
            else:
                try:
                    if opp == "==":
                        if d[ 1 ][ field ] == value:
                            tmp[ d[ 0 ] ] = d[ 1 ]

                    elif opp == "!=":
                        if d[ 1 ][ field ] != value:
                            tmp[ d[ 0 ] ] = d[ 1 ]

                except:
                    continue

        return tmp    
