#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import absolute_import
import hashlib
import sys
import time
import urllib
import urlparse

import cStringIO as StringIO
import psycopg2, psycopg2.extras



argparse_help = """Path to the PostgreSQL PiAnoS 4 database. This path has the
form 'pgsql://username:password@host:port/dbname'. You can use the short syntax
'/dbname', or omit some fields (e.g. pgsql://example.com/pianostest), to use the
defaults (username=pianostest, password=pianostest, host=localhost, port=5432)
"""

def parseConnectionString( string ):
	name = None
	user = 'pianostest'
	password = 'pianostest'
	hostname = 'localhost'
	port = 5432
	
	expectedScheme = 'pgsql'
	expectedFullScheme = expectedScheme + '://'
	
	if string[ :len( expectedFullScheme ) ] != expectedFullScheme:
		url = expectedFullScheme + string
	else:
		url = string
		
	res = urlparse.urlsplit( url )
	if res.scheme != expectedScheme:
		raise ValueError( "Invalid PostgreSQL url scheme in '%s'" % string )
		
	# Since urlparse does not recognize pgsql:// as a valid scheme, it will not parse correctly
	# the hostname:port/path?foo=bar#baz part. We thus replace pgsql:// by http:// to get the
	# correct parsed values.
	url = 'http' + url[ len( res.scheme ): ]
	res = urlparse.urlsplit( url )
	if len( res.path ):
		name = urllib.unquote( res.path.strip().strip( '/' ) )
	if res.username: 
		user = res.username
	if res.password: 
		password = res.password
	if res.hostname: 
		hostname = res.hostname
	if res.port:
		port = int( res.port )
	
	if name is None or user is None or password is None or hostname is None or port is None:
		raise ValueError( "Invalid PostgreSQL url: %s" % string )
		
	return hostname, port, name, user, password, 'pgsql://' + user + '@' + hostname + ':' + str( port ) + '/' + name
	
	
	
class Database( object ):
	############################################################################
	#
	# 	Administrativ functions
	#
	############################################################################
	def __init__( self, url, verbose = True, dont_register_admin = False, use_dict_cursors = True ):
		hostname, port, name, user, password, self.url = parseConnectionString( url )
		self.verbose = verbose
		if verbose:
			print "Connecting to '%s'" % self.url
		self.conn = psycopg2.connect( host = hostname, port = port, database = name, user = user, password = password, cursor_factory = psycopg2.extras.DictCursor if use_dict_cursors else None )
		c = self.conn.cursor()
		try:
			counts = []
			for x in [ 'object', 'image', 'exercise', 'person' ]:
				c.execute( "SELECT count(*) FROM " + x + ";" )
				counts.append( c.fetchone()[ 0 ] )
			if not dont_register_admin:
				c.execute( "SELECT set_config('unil.user',(SELECT person.id FROM person,membership,pgroup WHERE person.id = membership.person AND membership.pgroup = pgroup.id AND pgroup.tutors IS true ORDER BY person.id ASC LIMIT 1)::text,false);" )
		except psycopg2.ProgrammingError, err:
			raise ValueError( "'%s' is not a valid PiAnoS database (error reported: %s)" % ( self.url, str( err ) ) )
		if verbose:
			print "  Database OK (%d objects, %d images, %d exercises, %d users)" % tuple( counts )
		
	def cursor( self ):
		return self.conn.cusor()
	
	def commit( self ):
		"""
			Commit the queries to the database.
		"""
		self.conn.commit()
		
	def query( self, sql, *args, **kwargs ):
		"""
			Make a query on the database, and return the cursor.
		"""
		c = self.conn.cursor()
		c.execute( sql, *args, **kwargs )
		return c
	
	def query_fetchone( self, sql ):
		"""
			Make a query and return the first row.
		"""
		return self.query( sql ).fetchone()
	
	############################################################################
	# 	
	# 	Create exercises, images, ...
	# 	
	############################################################################
	
	def create_folder( self, name, personalof = None, childof = None, userid = None ):
		"""
			Check if the folder exists in the database. When existing, the folder.id is
			returned. Otherwise, the folder is created and the folder.id is returned.
		"""

		########################################################################
		
		if personalof != None:
			try:
				return self.query_fetchone( "SELECT folder.id FROM public.folder WHERE folder.personalof = %s;" % personalof )[ 0 ]
			except:
				id = self.query_fetchone( "INSERT INTO folder ( name, personalof ) VALUES ( '%s', %s ) RETURNING id;" % ( name, personalof ) )[ 0 ]
		
		else:
			vars = "name"
			content = "'%s'" % name
		
			if childof != None:
				vars += ", childof"
				content += ", %s" % childof
		
			try:
				sql = "SELECT folder.id FROM public.folder WHERE folder.name = '%s'" % name
				
				if childof != None:
					sql += " AND folder.childof = %s" % childof
				
				sql += ";"
				
				return self.query_fetchone( sql )[ 0 ]
			except:
				id = self.query_fetchone( "INSERT INTO folder ( %s ) VALUES ( %s ) RETURNING id;" % ( vars, content ) )[ 0 ]
		
		########################################################################
		
		if personalof != None:
			self.query_fetchone( "INSERT INTO object ( creator ) VALUES ( %s ) RETURNING id;" % personalof )[ 0 ]
			self.query( "INSERT INTO security ( object, trustee ) VALUES ( %s, %s );" % ( id, personalof ) )
		else:
			self.query_fetchone( "INSERT INTO object ( creator ) VALUES ( %s ) RETURNING id;" % 3 )[ 0 ]
			
		########################################################################
		
		return id
	
	# Image is a PIL.Image instance, 'type' is the short text ID from the imagetype table (e.g. 'MARK' or 'PRINT')
	def create_image( self, type, image, resolution, name = None ):
		c = self.conn.cursor()
		buf = StringIO.StringIO()
		image.save( buf, 'PNG' )
		
		h = hashlib.md5( image.tobytes() ).hexdigest()
		try:
			return self.query_fetchone( "SELECT id FROM image WHERE checksum = '%s'" % h )[ 0 ]
		except:
			try:
				filename = image.filename
			except:
				filename = name + "-" + type
				
			c.execute( """INSERT INTO image(type,name,mimetype,resolution,width,height,data,checksum)
				VALUES ((SELECT id FROM imagetype WHERE short=%s),%s,%s,%s,%s,%s,%s,%s) RETURNING id;""",
				( type, filename, 'image/png', resolution, image.size[ 0 ], image.size[ 1 ], memoryview( buf.getvalue() ), h ) )
			
			return c.fetchone()[ 0 ]
	
	# mark and _print must be PIL Images
	def create_exercise( self, folder_id, name, description, mark, mark_resolution, _print, _print_resolution, markname = None, _printname = None ):
		if markname == None:
			markname = name
		if _printname == None:
			_printname = name
		
		mark_id = self.create_image( 'MARK', mark, mark_resolution, markname )
		print_id = self.create_image( 'PRINT', _print, _print_resolution, _printname )
		c = self.conn.cursor()
		c.execute( "INSERT INTO exercise(name, description, folder, mark, print) VALUES (%s,%s,%s,%s,%s) RETURNING id;", ( name, description, folder_id, mark_id, print_id ) )
		return c.fetchone()[ 0 ]
	
	# mark and _print must be ids
	def create_exerciseById( self, folder_id, name, description, mark_id, print_id ):
		c = self.conn.cursor()
		c.execute( "INSERT INTO exercise(name, description, folder, mark, print) VALUES (%s,%s,%s,%s,%s) RETURNING id;", ( name, description, folder_id, mark_id, print_id ) )
		return c.fetchone()[ 0 ]
	
	# create group
	def create_group( self, name, tutor = False ):
		c = self.conn.cursor()
		c.execute( "INSERT INTO pgroup(name,tutors) VALUES (%s,%s) RETURNING id;", ( name, tutor, ) )
		return c.fetchone()[ 0 ]
	
	# create user
	def create_user( self, username, password, groupid, name = None, email = "", active = "true", suitability_approach = "2", must_change_pwd = "false" ):
		if name == None:
			name = username
		
		uid = "%s:%s" % ( username, password )
		h = hashlib.md5( uid ).hexdigest()
		
		id = self.query( "INSERT INTO person (username,name,password,email,active,language,suitability_approach,must_change_pwd) VALUES ('%s','%s','%s','%s','%s','1','%s','%s') RETURNING id;" % ( username, name, h, email, active, suitability_approach, must_change_pwd ) )
		id = id.fetchone()[ 0 ]
		
		self.query( "INSERT INTO membership (person,pgroup) VALUES ('%s','%s')" % ( id, groupid ) )
		
		return username + "\t" + password
	
	############################################################################
	# 
	# 	Get information
	# 
	############################################################################
	
	# get user id
	def get_userID( self, username ):
		return self.query_fetchone( "SELECT id FROM person WHERE username SIMILAR TO '%s';" % username )[ 0 ]
	
	def get_personalFolderOf( self, uid ):
		if type( uid ) == str:
			uid = self.get_userID( uid )
			
		return self.query_fetchone( "SELECT id FROM folder WHERE personalof = %s;" % uid )[ 0 ]
		
	# get image informations
	def get_image_info( self, image_id ):
		return self.query_fetchone( "SELECT width, height, resolution FROM image WHERE id = %s;" % image_id )
	
	# get exercise id
	def get_exercise_id( self, name ):
		return self.query_fetchone( "SELECT id FROM exercise WHERE name = '%s';" % name )[ 0 ]
	
	# get images for an exercise
	def get_images_exercise( self, ex_id ):
		return self.query_fetchone( "SELECT mark, print FROM exercise WHERE id = '%s';" % ex_id )
	
	# set security
	def assign_exerciseToGroup( self, exercise, group ):
		c = self.conn.cursor()
		c.execute( "INSERT INTO security(object,trustee) VALUES (%s,%s);", ( exercise, group, ) )
	
	############################################################################
	# 
	# 	Get exercises
	# 
	############################################################################
	
	# return all assignments, optionally filtered by exercise name, user name or limited in number
	def collect_assignments( self, exname = None, username = None, limit = None ):
		sql = '''SELECT assignment.id,
			object.createdon,
			getAssignmentStatus(assignment.id) AS status,
			person.id AS userid,person.username AS username,
			exercise.id AS exid, exercise.name AS exname,
			exercise.mark AS mark_image, exercise.print AS print_image
		FROM assignment
		JOIN object ON assignment.id = object.id 
		JOIN person ON object.creator = person.id
		JOIN exercise ON exercise.id = assignment.exercise
		WHERE assignment.ispreview IS false
		'''
		replacements = []
		
		filters = []
		if username:
			sql += " AND person.username SIMILAR TO %s"
			replacements.append( username )
			filters.append( ( 'user', username ) )
		if exname:
			sql += " AND exercise.name SIMILAR TO %s"
			replacements.append( exname )
			filters.append( ( 'exercise', exname ) )
		
		sql += " ORDER BY exid,userid,createdon DESC"

		if self.verbose:
			if len( filters ):
				print "Searching in results filtered by", ', '.join( [ x + '=' + y for x, y in filters ] )
			else:
				print "Searching in all results"
		
		if limit:
			if self.verbose:
				print "Limiting the result set to " + limit + " users"
			sql += " LIMIT " + str( int( limit ) ) + ";"
			
		c = self.conn.cursor()
		res = c.execute( sql, tuple( replacements ) )
		if self.verbose:
			print "Found", c.rowcount, "assignments"
		
		return c
	
	############################################################################
	# 
	# 	Set
	# 
	############################################################################
	
	def set_security( self, exid, userid ):
		if self.query_fetchone( "SELECT count( security.* ) FROM security WHERE object = %s AND trustee = %s" % ( exid, userid ) )[ 0 ] == 0:
			return self.query( "INSERT INTO security ( object, trustee ) VALUES ( %s, %s );" % ( exid, userid ) )
	
	############################################################################
	#
	# 	Cleaning
	#
	############################################################################
	
	def delete_folder( self, folderid ):
		try:
			for exid in self.query( "SELECT id FROM exercise WHERE folder = %s;" % folderid ).fetchall():
				print "ex", exid
				self.delete_exercise( exid[ 0 ] )
		except:
			pass
	
		try:
			for subfolderid in self.query( "SELECT id FROM folder WHERE childof = %s;" % folderid ).fetchall():
				print "subfolder", subfolderid
				self.delete_folder( subfolderid[ 0 ] )
		
		finally:
			self.delete_folder_only( folderid )
	
	def delete_folder_only( self, folderid ):
		self.query( "DELETE FROM security WHERE object = %s;" % folderid )
		self.query( "DELETE FROM folder WHERE id = %s;" % folderid )
		self.query( "DELETE FROM object WHERE id = %s;" % folderid )
	
	def delete_exercise( self, exid ):
		try:
			assid = self.query_fetchone( "SELECT id FROM assignment WHERE exercise = %s;" % exid )[ 0 ]
			
			self.query( "DELETE FROM minutia WHERE assignment = %s;" % assid )
			self.query( "DELETE FROM observation WHERE assignment = %s;" % assid )
			self.query( "DELETE FROM annotation WHERE assignment = %s;" % assid )
			self.query( "DELETE FROM result WHERE assignment = %s;" % assid )
			self.query( "DELETE FROM resulthistory WHERE assignment = %s;" % assid )
			self.query( "DELETE FROM assignment WHERE id = %s;" % assid )
			self.query( "DELETE FROM object WHERE id = %s;" % assid )
		
		finally:
			self.query( "DELETE FROM security WHERE object = %s;" % exid )
			self.query( "DELETE FROM exercise WHERE id = %s;" % exid )
			self.query( "DELETE FROM object WHERE id = %s;" % exid )
	
	def delete_unused_images( self ):
		sql = """
			SELECT
			    imagev.id, array_agg(ex.id) AS ex_ids
			
			FROM
			    person, object, imagev
			
			LEFT OUTER JOIN
			    exercise AS ex
			    ON ( ex.mark = imagev.id OR ex.print = imagev.id )
			
			WHERE
			    imagev.typeid <> ( SELECT id FROM imagetype WHERE short = E'THUMBNAIL' ) AND
			    person.id = object.creator AND
			    object.id = imagev.id
			
			GROUP BY imagev.id
			
			ORDER BY imagev.id ASC;
		"""
		
		for id in [ id for id, exs in self.query( sql ).fetchall() if exs == [None] ]:
			sql = "DELETE FROM security WHERE object = %s;" % id
			self.query( sql )
			
			sql = "DELETE FROM image WHERE id = %s;" % id
			self.query( sql )
			
			sql = "DELETE FROM object WHERE id = %s;" % id
			self.query( sql )
			
		self.commit()
	
	############################################################################
	#
	# 	Other
	#
	############################################################################
	
	def close( self ):
		self.conn.close()
	
	############################################################################
	#
	# 	Context manager
	#
	############################################################################
	
	def __enter__( self ):
		return self
	
	def __exit__ ( self, type, value, tb ):
		if tb is None:
			self.commit()
		
		self.close()
		
class Timer( object ):
	
	def __init__( self, total, max_itr = sys.maxint ):
		self.total = total
		self.ftotal = float( self.total )
		self.current = 0
		self.steps_before_next_estimation = 1
		self.estimation_steps = 0
		self.last_estimation_time = 0
		self.max_itr = max_itr
		
	def start( self ):
		self.start_time = time.time()
		
	def next( self, force_estimation = False ):
		self.current += 1
		self.estimation_steps += 1
		progress = float( self.current ) / self.ftotal * 100.0
		eta = None
		if force_estimation or self.estimation_steps >= self.steps_before_next_estimation:
			now = time.time()
			elapsed = now - self.start_time
			if elapsed > 0.0:
				speed = float( self.current ) / elapsed
				elapsed_last_check = now - self.last_estimation_time
				# print "Averaged last %d iterations (%f secs total, %f last check) to %fitr/sec" % (self.estimation_steps, elapsed, elapsed_last_check, speed)
				if elapsed_last_check < 0.5:
					self.steps_before_next_estimation = min( self.max_itr, 2 * self.steps_before_next_estimation )
				elif elapsed_last_check > 2.0 and self.steps_before_next_estimation >= 2:
					self.steps_before_next_estimation /= 2
				self.estimation_steps = 0
				self.last_estimation_time = now
				remaining = self.total - self.current
				if remaining < 0:	
					eta = 0.0
				else:
					eta = remaining / speed
		return progress, eta
			
	def end( self ):
		self.end_time = time.time()
		return self.end_time - self.start_time
		
